#include "fftk.h"

void FFTK::Init(string basis, int Nx, int Ny, int Nz, int num_p_rows)
{
	this->basis=basis;
	this->Nx=Nx;
	this->Ny=Ny;
	this->Nz=Nz;
	this->num_p_rows=num_p_rows;

	if (basis[0] == 'F' || basis[0] == 'S') {
		Fx=Nx;
		Ix=Nx;
		Rx=Nx;
	}
	else {
		cerr << "basis[0] can be S or F";
		MPI_Abort(MPI_COMM_WORLD, 1);
	}


	if (basis[1] == 'F' || basis[1] == 'S') {
		Fy=Ny;
		Iy=Ny;
		Ry=Ny;
	}
	else {
		cerr << "basis[1] can be S or F";
		MPI_Abort(MPI_COMM_WORLD, 1);
	}

	if (basis[2] == 'F') {
		Fz=Nz/2+1;
		Iz=Nz/2+1;
		Rz=Nz+2;

		Fz_unit=2;
		Iz_unit=2;
		Rz_unit=1;
	}
	else if (basis[2] == 'S') {
		Fz=Nz/2;
		Iz=Nz/2;
		Rz=Nz;

		Fz_unit=2;
		Iz_unit=2;
		Rz_unit=1;
	}
	else {
		cerr << "basis[] can be S or F";
		MPI_Abort(MPI_COMM_WORLD, 1);
	}


	set_vars();
	init_transpose();
	init_transform();
	init_timers();
}

void FFTK::set_vars() {
	MPI_Comm_rank(MPI_COMM_WORLD, &my_id);
	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
	
	num_p_cols=numprocs/num_p_rows;

	MPI_Comm_split(MPI_COMM_WORLD, my_id/num_p_cols, 0, &MPI_COMM_ROW);
	MPI_Comm_split(MPI_COMM_WORLD, my_id%num_p_cols, 0, &MPI_COMM_COL);

	MPI_Comm_rank(MPI_COMM_ROW, &my_col_id);
	MPI_Comm_rank(MPI_COMM_COL, &my_row_id);

	if (Ny>1) {
		maxfx=Fx;
		maxfy=Fy/num_p_cols;
		maxfz=Fz/num_p_rows;

		maxix=Ix/num_p_cols;
		maxiy=Iy;
		maxiz=Iz/num_p_rows;

		maxrx=Rx/num_p_cols;
		maxry=Ry/num_p_rows;
		maxrz=Rz;
	}
	else {
		maxfx=Fx;
		maxfy=1;
		maxfz=Fz/num_p_cols;

		maxix=Ix/num_p_cols;
		maxiy=1;
		maxiz=Iz/num_p_rows;

		maxrx=Rx/num_p_cols;
		maxry=1;
		maxrz=Rz;	
	}


	if (Ny>1) {
		fx_start=0;
		fy_start=maxfy*my_col_id;
		fz_start=maxfz*my_row_id;

		ix_start=maxix*my_col_id;
		iy_start=0;
		iz_start=maxiz*my_row_id;

		rx_start=maxrx*my_col_id;
		ry_start=maxry*my_row_id;
		rz_start=0;
	}
	else {
		fx_start=0;
		fy_start=0;
		fz_start=maxfz*my_col_id;

		ix_start=-1;
		iy_start=-1;
		iz_start=-1;

		rx_start=maxrx*my_col_id;
		ry_start=0;
		rz_start=0;	
	}

	//Set Array size
	FA_shape=shape(maxfx, maxfy, maxfz);
	IA_shape=shape(maxix, maxiy, maxiz);
	RA_shape=shape(maxrx, maxry, maxrz);

	//FA.resize(FA_shape);
	IA.resize(IA_shape);
	RA.resize(RA_shape);

	fft_plan_Fourier_forward_x=NULL;
	fft_plan_Fourier_inverse_x=NULL;
	fft_plan_Fourier_forward_y=NULL;
	fft_plan_Fourier_inverse_y=NULL;
	fft_plan_Fourier_forward_z=NULL;
	fft_plan_Fourier_inverse_z=NULL;

	fft_plan_Sine_forward_x=NULL;
	fft_plan_Sine_inverse_x=NULL;
	fft_plan_Sine_forward_y=NULL;
	fft_plan_Sine_inverse_y=NULL;
	fft_plan_Sine_forward_z=NULL;
	fft_plan_Sine_inverse_z=NULL;

	fft_plan_Cosine_forward_x=NULL;
	fft_plan_Cosine_inverse_x=NULL;
	fft_plan_Cosine_forward_y=NULL;
	fft_plan_Cosine_inverse_y=NULL;
	fft_plan_Cosine_forward_z=NULL;
	fft_plan_Cosine_inverse_z=NULL;

}


void FFTK::init_transpose()
{
	//To revise
	if (Ny>1)
		buffer_Y = new Real[size_t(Nx/num_p_cols)*size_t(Ny/num_p_rows)*size_t(Nz+2)];

	buffer_Z = new Real[size_t(Nx/num_p_cols)*size_t(Ny/num_p_rows)*size_t(Nz+2)];

	plane_size_YX = Iz_unit*(Iz/num_p_rows)*Iy;
	copy_size_YX  = Iz_unit*(Iz/num_p_rows)*(Iy/num_p_cols);
	block_size_YX = Iz_unit*(Iz/num_p_rows)*(Iy/num_p_cols)*(Ix/num_p_cols);

	plane_size_YZ = Iz_unit*(Iz/num_p_rows)*Iy;
	copy_size_YZ  = Iz_unit*(Iz/num_p_rows)*(Iy/num_p_rows);
	block_size_YZ = Iz_unit*(Iz/num_p_rows)*(Iy/num_p_rows)*(Ix/num_p_cols);
}

void FFTK::init_transform(){
	//Initialize plans
	int Nx_dims[]={Nx};
	int Ny_dims[]={Ny};
	int Nz_dims[]={Nz};

	fftw_r2r_kind kind[1];

	normalizing_factor=1;


#ifdef FFTK_ENABLE_THREADS
	fftw_init_threads();
	fftw_plan_with_nthreads(omp_get_max_threads());
#endif

 	//X transforms
 	if ( Nx>1 ) {
		if (basis[0] == 'F') {
			fft_plan_Fourier_forward_x = FFTW_PLAN_MANY_DFT(1, Nx_dims, maxfy*maxfz,
				reinterpret_cast<FFTW_Complex*>(IA.data()), NULL,
				maxfy*maxfz, 1,
				reinterpret_cast<FFTW_Complex*>(IA.data()), NULL,
				maxfy*maxfz, 1,
				FFTW_FORWARD, FFTW_PLAN_FLAG);

			fft_plan_Fourier_inverse_x = FFTW_PLAN_MANY_DFT(1, Nx_dims, maxfy*maxfz,
				reinterpret_cast<FFTW_Complex*>(IA.data()), NULL,
				maxfy*maxfz, 1,
				reinterpret_cast<FFTW_Complex*>(IA.data()), NULL,
				maxfy*maxfz, 1,
				FFTW_BACKWARD, FFTW_PLAN_FLAG);

			normalizing_factor/=Nx;

			if (fft_plan_Fourier_forward_x == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Fourier_forward_x'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Fourier_inverse_x == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Fourier_inverse_x'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
		}
		else if (basis[0] == 'S') {
			kind[0]=FFTW_RODFT10;
			fft_plan_Sine_forward_x = FFTW_PLAN_MANY_R2R(1, Nx_dims, Fz_unit*maxfz*maxfy,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Fz_unit*maxfz*maxfy, 1,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Fz_unit*maxfz*maxfy, 1,
			    kind, FFTW_PLAN_FLAG);

			kind[0]=FFTW_RODFT01;
			fft_plan_Sine_inverse_x = FFTW_PLAN_MANY_R2R(1, Nx_dims, Fz_unit*maxfz*maxfy,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Fz_unit*maxfz*maxfy, 1,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Fz_unit*maxfz*maxfy, 1,
			    kind, FFTW_PLAN_FLAG);

			kind[0]=FFTW_REDFT10;
			fft_plan_Cosine_forward_x = FFTW_PLAN_MANY_R2R(1, Nx_dims, Fz_unit*maxfz*maxfy,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Fz_unit*maxfz*maxfy, 1,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Fz_unit*maxfz*maxfy, 1,
			    kind, FFTW_PLAN_FLAG);

			kind[0]=FFTW_REDFT01;
			fft_plan_Cosine_inverse_x = FFTW_PLAN_MANY_R2R(1, Nx_dims, Fz_unit*maxfz*maxfy,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Fz_unit*maxfz*maxfy, 1,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Fz_unit*maxfz*maxfy, 1,
			    kind, FFTW_PLAN_FLAG);

			normalizing_factor/=2*Nx;

			if (fft_plan_Sine_forward_x == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Sine_forward_x'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Sine_inverse_x == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Sine_inverse_x'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Cosine_forward_x == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Cosine_forward_x'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Cosine_inverse_x == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Cosine_inverse_x'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
		}
		else {
			if (my_id==0) cerr << "FFTK::init_transform() - invalid basis" << endl;
			MPI_Abort(MPI_COMM_WORLD, 1);
		}
	}



	//Y transforms
	if ( Ny>1 ) {
		if (basis[1] == 'F') {
			fft_plan_Fourier_forward_y = FFTW_PLAN_MANY_DFT(1, Ny_dims, maxiz,
				reinterpret_cast<FFTW_Complex*>(IA.data()), NULL,
				maxiz, 1,
				reinterpret_cast<FFTW_Complex*>(IA.data()), NULL,
				maxiz, 1,
				FFTW_FORWARD, FFTW_PLAN_FLAG);

			fft_plan_Fourier_inverse_y = FFTW_PLAN_MANY_DFT(1, Ny_dims, maxiz,
				reinterpret_cast<FFTW_Complex*>(IA.data()), NULL,
				maxiz, 1,
				reinterpret_cast<FFTW_Complex*>(IA.data()), NULL,
				maxiz, 1,
				FFTW_BACKWARD, FFTW_PLAN_FLAG);

			normalizing_factor/=Ny;

			if (fft_plan_Fourier_forward_y == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Fourier_forward_y'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Fourier_inverse_y == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Fourier_inverse_y'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
		}
		else if (basis[1] == 'S') {
			kind[0]=FFTW_RODFT10;
			fft_plan_Sine_forward_y = FFTW_PLAN_MANY_R2R(1, Ny_dims, Iz_unit*maxiz,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Iz_unit*maxiz, 1,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Iz_unit*maxiz, 1,
			    kind, FFTW_PLAN_FLAG);

			kind[0]=FFTW_RODFT01;
			fft_plan_Sine_inverse_y = FFTW_PLAN_MANY_R2R(1, Ny_dims, Iz_unit*maxiz,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Iz_unit*maxiz, 1,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Iz_unit*maxiz, 1,
			    kind, FFTW_PLAN_FLAG);

			kind[0]=FFTW_REDFT10;
			fft_plan_Cosine_forward_y = FFTW_PLAN_MANY_R2R(1, Ny_dims, Iz_unit*maxiz,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Iz_unit*maxiz, 1,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Iz_unit*maxiz, 1,
			    kind, FFTW_PLAN_FLAG);

			kind[0]=FFTW_REDFT01;
			fft_plan_Cosine_inverse_y = FFTW_PLAN_MANY_R2R(1, Ny_dims, Iz_unit*maxiz,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Iz_unit*maxiz, 1,
			    reinterpret_cast<Real*>(IA.data()), NULL,
			    Iz_unit*maxiz, 1,
			    kind, FFTW_PLAN_FLAG);

			normalizing_factor/=2*Ny;

			if (fft_plan_Sine_forward_y == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Sine_forward_y'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Sine_inverse_y == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Sine_inverse_y'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Cosine_forward_y == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Cosine_forward_y'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Cosine_inverse_y == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Cosine_inverse_y'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
		}
		else {
			if (my_id==0) cerr << "FFTK::init_transform() - invalid basis" << endl;
			MPI_Abort(MPI_COMM_WORLD, 1);
		}
	}

	//Z transforms
	if ( Nz>1) {
		if (basis[2] == 'F') {
			fft_plan_Fourier_forward_z = FFTW_PLAN_MANY_DFT_R2C(1, Nz_dims, maxrx*maxry,
				reinterpret_cast<Real*>(RA.data()), NULL,
				1, Nz+2,
				reinterpret_cast<FFTW_Complex*>(RA.data()), NULL,
				1, Nz/2+1,
				FFTW_PLAN_FLAG);

			fft_plan_Fourier_inverse_z = FFTW_PLAN_MANY_DFT_C2R(1, Nz_dims, maxrx*maxry,
				reinterpret_cast<FFTW_Complex*>(RA.data()), NULL,
				1, Nz/2+1,
				reinterpret_cast<Real*>(RA.data()), NULL,
				1, Nz+2,
				FFTW_PLAN_FLAG);

			normalizing_factor/=Nz;

			if (fft_plan_Fourier_forward_z == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Fourier_forward_z'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Fourier_inverse_z == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Fourier_inverse_z'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
		}
		else if (basis[2] == 'S') {
			kind[0]=FFTW_RODFT10;
			fft_plan_Sine_forward_z = FFTW_PLAN_MANY_R2R(1, Nz_dims, maxrx*maxry,
			    reinterpret_cast<Real*>(RA.data()), NULL,
			    1, Nz,
			    reinterpret_cast<Real*>(RA.data()), NULL,
			    1, Nz,
			    kind, FFTW_PLAN_FLAG);

			kind[0]=FFTW_RODFT01;
			fft_plan_Sine_inverse_z = FFTW_PLAN_MANY_R2R(1, Nz_dims, maxrx*maxry,
			    reinterpret_cast<Real*>(RA.data()), NULL,
			    1, Nz,
			    reinterpret_cast<Real*>(RA.data()), NULL,
			    1, Nz,
			    kind, FFTW_PLAN_FLAG);

			kind[0]=FFTW_REDFT10;
			fft_plan_Cosine_forward_z = FFTW_PLAN_MANY_R2R(1, Nz_dims, maxrx*maxry,
			    reinterpret_cast<Real*>(RA.data()), NULL,
			    1, Nz,
			    reinterpret_cast<Real*>(RA.data()), NULL,
			    1, Nz,
			    kind, FFTW_PLAN_FLAG);

			kind[0]=FFTW_REDFT01;
			fft_plan_Cosine_inverse_z = FFTW_PLAN_MANY_R2R(1, Nz_dims, maxrx*maxry,
			    reinterpret_cast<Real*>(RA.data()), NULL,
			    1, Nz,
			    reinterpret_cast<Real*>(RA.data()), NULL,
			    1, Nz,
			    kind, FFTW_PLAN_FLAG);

			normalizing_factor/=2*Nz;

			if (fft_plan_Sine_forward_z == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Sine_forward_z'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Sine_inverse_z == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Sine_inverse_z'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Cosine_forward_z == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Cosine_forward_z'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
			if (fft_plan_Cosine_inverse_z == NULL){
				if (my_id==0) cerr << "failed to initialize 'fft_plan_Cosine_inverse_z'" << endl;
				MPI_Abort(MPI_COMM_WORLD, 1);
			}
		}
		else {
			if (my_id==0) cerr << "FFTK::init_transform() - invalid basis" << endl;
			MPI_Abort(MPI_COMM_WORLD, 1);
		}
	}
}


void FFTK::init_timers(){
	timer = new double[22];
	for (int i=0; i<22; i++)
		timer[i]=0;
}


void FFTK::Forward_transform(string basis_option, Array<Real,2> Ar, Array<Complex,2> A)
{
	// timer[0]-=MPI_Wtime();
	if ( (basis_option[2] == 'F') && (fft_plan_Fourier_forward_z != NULL) ) {
		FFTW_EXECUTE_DFT_R2C(fft_plan_Fourier_forward_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<FFTW_Complex*>(Ar.data()));
	}
	else if ( (basis_option[2] == 'S') && (fft_plan_Sine_forward_z != NULL) ) {
		FFTW_EXECUTE_R2R(fft_plan_Sine_forward_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
		ArrayShiftRight(Ar, 'Z');
	}
	else if ( (basis_option[2] == 'C') && (fft_plan_Cosine_forward_z != NULL) )
		FFTW_EXECUTE_R2R(fft_plan_Cosine_forward_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
	else {
		if (my_id==0) cerr << "FFTK::forward_transform - Invalid basis_option '" << basis_option << "' or plan z not initialized" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[0]+=MPI_Wtime();

	// timer[1]-=MPI_Wtime();
	Transpose(Ar, A);
	// timer[1]+=MPI_Wtime();


	// timer[4]-=MPI_Wtime();
	if ( (basis_option[0] == 'F') && (fft_plan_Fourier_forward_x != NULL) )
		FFTW_EXECUTE_DFT(fft_plan_Fourier_forward_x, reinterpret_cast<FFTW_Complex*>(A.data()), reinterpret_cast<FFTW_Complex*>(A.data()));
	else if ( (basis_option[0] == 'S') && (fft_plan_Sine_forward_x != NULL) ) {
		FFTW_EXECUTE_R2R(fft_plan_Sine_forward_x, reinterpret_cast<Real*>(A.data()), reinterpret_cast<Real*>(A.data()));
		ArrayShiftRight(A, 'X');
	}
	else if ( (basis_option[0] == 'C') && (fft_plan_Cosine_forward_x != NULL) )
		FFTW_EXECUTE_R2R(fft_plan_Cosine_forward_x, reinterpret_cast<Real*>(A.data()), reinterpret_cast<Real*>(A.data()));
	else {
		if (my_id==0) cerr << "FFTK::forward_transform - Invalid basis_option '" << basis_option << "' or plan x not initialized" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[4]+=MPI_Wtime();

	// timer[5]-=MPI_Wtime();
	Normalize(A);
	// timer[5]+=MPI_Wtime();
}


void FFTK::Inverse_transform(string basis_option, Array<Complex,2> A, Array<Real,2> Ar)
{
	// timer[6]-=MPI_Wtime();
	if ( (basis_option[0] == 'F') && (fft_plan_Fourier_inverse_x != NULL) )
		FFTW_EXECUTE_DFT(fft_plan_Fourier_inverse_x, reinterpret_cast<FFTW_Complex*>(A.data()), reinterpret_cast<FFTW_Complex*>(A.data()));
	else if ( (basis_option[0] == 'S') && (fft_plan_Sine_inverse_x != NULL) ) {
		ArrayShiftLeft(A, 'X');
		FFTW_EXECUTE_R2R(fft_plan_Sine_inverse_x, reinterpret_cast<Real*>(A.data()), reinterpret_cast<Real*>(A.data()));
	}
	else if ( (basis_option[0] == 'C') && (fft_plan_Cosine_inverse_x != NULL) )
		FFTW_EXECUTE_R2R(fft_plan_Cosine_inverse_x, reinterpret_cast<Real*>(A.data()), reinterpret_cast<Real*>(A.data()));
	else {
		if (my_id==0) cerr << "FFTK::inverse_transform - Invalid basis_option '" << basis_option << "'" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[6]+=MPI_Wtime();


	// timer[7]-=MPI_Wtime();
	Transpose(A, Ar);
	// timer[7]+=MPI_Wtime();


	// timer[10]-=MPI_Wtime();
	if ( (basis_option[2] == 'F') && (fft_plan_Fourier_inverse_z != NULL) )
		FFTW_EXECUTE_DFT_C2R(fft_plan_Fourier_inverse_z, reinterpret_cast<FFTW_Complex*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
	else if ( (basis_option[2] == 'S') && (fft_plan_Sine_inverse_z != NULL) ){
		ArrayShiftLeft(Ar, 'Z');
		FFTW_EXECUTE_R2R(fft_plan_Sine_inverse_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
	}
	else if ( (basis_option[2] == 'C') && (fft_plan_Cosine_inverse_z != NULL) )
		FFTW_EXECUTE_R2R(fft_plan_Cosine_inverse_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
	else {
		if (my_id==0) cerr << "FFTK::forward_transform - Invalid basis_option '" << basis_option << "'" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[10]+=MPI_Wtime();
}



void FFTK::Forward_transform(string basis_option, Array<Real,3> Ar, Array<Complex,3> A)
{
	// timer[0]-=MPI_Wtime();
	if ( (basis_option[2] == 'F') && (fft_plan_Fourier_forward_z != NULL) )
		FFTW_EXECUTE_DFT_R2C(fft_plan_Fourier_forward_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<FFTW_Complex*>(Ar.data()));
	else if ( (basis_option[2] == 'S') && (fft_plan_Sine_forward_z != NULL) ) {
		FFTW_EXECUTE_R2R(fft_plan_Sine_forward_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
		ArrayShiftRight(Ar, 'Z');
	}
	else if ( (basis_option[2] == 'C') && (fft_plan_Cosine_forward_z != NULL) )
		FFTW_EXECUTE_R2R(fft_plan_Cosine_forward_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
	else {
		if (my_id==0) cerr << "FFTK::forward_transform - Invalid basis_option '" << basis_option << "' or plan z not initialized" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[0]+=MPI_Wtime();


	// timer[1]-=MPI_Wtime();
	Transpose_ZY(Ar, IA);
	// timer[1]+=MPI_Wtime();


	// timer[2]-=MPI_Wtime();
	if ( (basis_option[1] == 'F') && (fft_plan_Fourier_forward_y != NULL) )
		for (int ix = 0; ix < maxix; ix++)
			FFTW_EXECUTE_DFT(fft_plan_Fourier_forward_y, reinterpret_cast<FFTW_Complex*>(IA(ix,Range::all(),Range::all()).data()), reinterpret_cast<FFTW_Complex*>(IA(ix,Range::all(),Range::all()).data()));
	else if ( (basis_option[1] == 'S') && (fft_plan_Sine_forward_y != NULL) ) {
		for (int ix = 0; ix < maxix; ix++)
			FFTW_EXECUTE_R2R(fft_plan_Sine_forward_y, reinterpret_cast<Real*>(IA(ix,Range::all(),Range::all()).data()), reinterpret_cast<Real*>(IA(ix,Range::all(),Range::all()).data()));
		ArrayShiftRight(IA, 'Y');
	}
	else if ( (basis_option[1] == 'C') && (fft_plan_Cosine_forward_y != NULL) )
		for (int ix = 0; ix < maxix; ix++)
			FFTW_EXECUTE_R2R(fft_plan_Cosine_forward_y, reinterpret_cast<Real*>(IA(ix,Range::all(),Range::all()).data()), reinterpret_cast<Real*>(IA(ix,Range::all(),Range::all()).data()));
	else {
		if (my_id==0) cerr << "FFTK::forward_transform - Invalid basis_option '" << basis_option << "' or plan y not initialized" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[2]+=MPI_Wtime();


	// timer[3]-=MPI_Wtime();
	Transpose_YX(IA, A);
	// timer[3]+=MPI_Wtime();


	// timer[4]-=MPI_Wtime();
	if ( (basis_option[0] == 'F') && (fft_plan_Fourier_forward_x != NULL) )
		FFTW_EXECUTE_DFT(fft_plan_Fourier_forward_x, reinterpret_cast<FFTW_Complex*>(A.data()), reinterpret_cast<FFTW_Complex*>(A.data()));
	else if ( (basis_option[0] == 'S') && (fft_plan_Sine_forward_x != NULL) ) {
		FFTW_EXECUTE_R2R(fft_plan_Sine_forward_x, reinterpret_cast<Real*>(A.data()), reinterpret_cast<Real*>(A.data()));
		ArrayShiftRight(A, 'X');
	}
	else if ( (basis_option[0] == 'C') && (fft_plan_Cosine_forward_x != NULL) )
		FFTW_EXECUTE_R2R(fft_plan_Cosine_forward_x, reinterpret_cast<Real*>(A.data()), reinterpret_cast<Real*>(A.data()));
	else {
		if (my_id==0) cerr << "FFTK::forward_transform - Invalid basis_option '" << basis_option << "' or plan x not initialized" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[4]+=MPI_Wtime();

	// timer[5]-=MPI_Wtime();
	Normalize(A);
	// timer[5]+=MPI_Wtime();
}


void FFTK::Inverse_transform(string basis_option, Array<Complex,3> A, Array<Real,3> Ar) 
{
	// timer[6]-=MPI_Wtime();
	if ( (basis_option[0] == 'F') && (fft_plan_Fourier_inverse_x != NULL) )
		FFTW_EXECUTE_DFT(fft_plan_Fourier_inverse_x, reinterpret_cast<FFTW_Complex*>(A.data()), reinterpret_cast<FFTW_Complex*>(A.data()));
	else if ( (basis_option[0] == 'S') && (fft_plan_Sine_inverse_x != NULL) ) {
		ArrayShiftLeft(A, 'X');
		FFTW_EXECUTE_R2R(fft_plan_Sine_inverse_x, reinterpret_cast<Real*>(A.data()), reinterpret_cast<Real*>(A.data()));
	}
	else if ( (basis_option[0] == 'C') && (fft_plan_Cosine_inverse_x != NULL) )
		FFTW_EXECUTE_R2R(fft_plan_Cosine_inverse_x, reinterpret_cast<Real*>(A.data()), reinterpret_cast<Real*>(A.data()));
	else {
		if (my_id==0) cerr << "FFTK::inverse_transform - Invalid basis_option '" << basis_option << "'" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[6]+=MPI_Wtime();


	// timer[7]-=MPI_Wtime();
	Transpose_XY(A, IA);
	// timer[7]+=MPI_Wtime();


	// timer[8]-=MPI_Wtime();
	if ( (basis_option[1] == 'F') && (fft_plan_Fourier_inverse_y != NULL) )
		for (int ix = 0; ix < maxix; ix++)
			FFTW_EXECUTE_DFT(fft_plan_Fourier_inverse_y, reinterpret_cast<FFTW_Complex*>(IA(ix,Range::all(),Range::all()).data()), reinterpret_cast<FFTW_Complex*>(IA(ix,Range::all(),Range::all()).data()));
	else if ( (basis_option[1] == 'S') && (fft_plan_Sine_inverse_y != NULL) ){
		ArrayShiftLeft(IA, 'Y');
		for (int ix = 0; ix < maxix; ix++)
			FFTW_EXECUTE_R2R(fft_plan_Sine_inverse_y, reinterpret_cast<Real*>(IA(ix,Range::all(),Range::all()).data()), reinterpret_cast<Real*>(IA(ix,Range::all(),Range::all()).data()));
	}
	else if ( (basis_option[1] == 'C') && (fft_plan_Cosine_inverse_y != NULL) )
		for (int ix = 0; ix < maxix; ix++)
			FFTW_EXECUTE_R2R(fft_plan_Cosine_inverse_y, reinterpret_cast<Real*>(IA(ix,Range::all(),Range::all()).data()), reinterpret_cast<Real*>(IA(ix,Range::all(),Range::all()).data()));
	else {
		if (my_id==0) cerr << "FFTK::inverse_transform - Invalid basis_option '" << basis_option << "'" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[8]+=MPI_Wtime();


	// timer[9]-=MPI_Wtime();
	Transpose_YZ(IA, Ar);
	// timer[9]+=MPI_Wtime();


	// timer[10]-=MPI_Wtime();
	if ( (basis_option[2] == 'F') && (fft_plan_Fourier_inverse_z != NULL) )
		FFTW_EXECUTE_DFT_C2R(fft_plan_Fourier_inverse_z, reinterpret_cast<FFTW_Complex*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
	else if ( (basis_option[2] == 'S') && (fft_plan_Sine_inverse_z != NULL) ){
		ArrayShiftLeft(Ar, 'Z');
		FFTW_EXECUTE_R2R(fft_plan_Sine_inverse_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
	}
	else if ( (basis_option[2] == 'C') && (fft_plan_Cosine_inverse_z != NULL) )
		FFTW_EXECUTE_R2R(fft_plan_Cosine_inverse_z, reinterpret_cast<Real*>(Ar.data()), reinterpret_cast<Real*>(Ar.data()));
	else {
		if (my_id==0) cerr << "FFTK::forward_transform - Invalid basis_option '" << basis_option << "'" << endl;
		MPI_Abort(MPI_COMM_WORLD, 2);
	}
	// timer[10]+=MPI_Wtime();
}

void FFTK::Transpose(Array<Complex,2> FA, Array<Real,2> RA)
{
	if (num_p_cols>1) {
		Real *__restrict__ RA_data=reinterpret_cast<Real*>(RA.data());
		Real *__restrict__ FA_data=reinterpret_cast<Real*>(FA.data());

		MPI_Alltoall(FA_data, maxfz*Fz_unit*maxrx, MPI_Real, buffer_Z, maxfz*Fz_unit*maxrx, MPI_Real, MPI_COMM_ROW);

		for (size_t c=0; c<num_p_cols; c++)
			for (size_t rx=0; rx<maxrx; rx++)
				std::copy(buffer_Z + c*maxrx*maxfz*Fz_unit + rx*maxfz*Fz_unit, buffer_Z + c*maxrx*maxfz*Fz_unit + (rx*maxfz+1)*Fz_unit, RA_data + rx*Rz + c*maxfz*Fz_unit);

	}
	else
		RA=Array<Real,2>(reinterpret_cast<Real*>(FA.data()), RA.shape(), neverDeleteData);
}

void FFTK::Transpose(Array<Real,2> RA, Array<Complex,2> FA)
{
	if (num_p_cols>1) {
		Real *__restrict__ RA_data=reinterpret_cast<Real*>(RA.data());
		Real *__restrict__ FA_data=reinterpret_cast<Real*>(FA.data());

		for (size_t c=0; c<num_p_cols; c++)
			for (size_t rx=0; rx<maxrx; rx++)
				for (int i=0; i<maxfz*Fz_unit; i++)
					*(buffer_Z + c*maxrx*maxfz*Fz_unit + rx*maxfz*Fz_unit + i) = *(RA_data + rx*Rz + c*maxfz*Fz_unit+i);
		// 		std::copy(RA_data + rx*(Nz+2) + c*maxfz*Fz_unit, RA_data + rx*(Nz+2) + (c+1)*maxfx*Fz_unit, RA_data + c*maxrx*maxfz*Fz_unit + rx*maxfz*Fz_unit);

		MPI_Alltoall(buffer_Z, maxfz*Fz_unit*maxrx, MPI_Real, FA_data, maxfz*Fz_unit*maxrx, MPI_Real, MPI_COMM_ROW);
	}
	else
		FA=Array<Complex,2>(reinterpret_cast<Complex*>(RA.data()), FA.shape(), neverDeleteData);
}



void FFTK::Transpose(Array<Real,3> Ar, Array<Complex,3> A)
{
	Transpose_ZY(Ar,IA);
	Transpose_YX(IA,A);
}

void FFTK::Transpose(Array<Complex,3> A, Array<Real,3> Ar)
{
	Transpose_XY(A,IA);
	Transpose_YZ(IA,Ar);
}


void FFTK::Transpose_XY(Array<Complex,3> FA, Array<Complex,3> IA)
{

	if (num_p_cols>1) {
		Real *__restrict__ FA_data=reinterpret_cast<Real*>(FA.data());
		Real *__restrict__ IA_data=reinterpret_cast<Real*>(IA.data());

		// timer[11] -= MPI_Wtime();
		MPI_Alltoall(FA_data, block_size_YX, MPI_Real, buffer_Y, block_size_YX, MPI_Real, MPI_COMM_ROW);
		// timer[11] += MPI_Wtime();

		// timer[12] -= MPI_Wtime();
		for (int c=0; c<num_p_cols; c++)
			for (int ix=0; ix<maxix; ix++)
				std::copy(buffer_Y + c*block_size_YX + ix*copy_size_YX, buffer_Y + c*block_size_YX + (ix+1)*copy_size_YX, IA_data + ix*plane_size_YX + c*copy_size_YX);
		// timer[12] += MPI_Wtime();
	}
	else
		IA=Array<Complex,3>(reinterpret_cast<Complex*>(FA.data()), IA_shape, neverDeleteData);

}

void FFTK::Transpose_YX(Array<Complex,3> IA, Array<Complex,3> FA)
{
	if (num_p_cols>1) {
		Real *__restrict__ IA_data=reinterpret_cast<Real*>(IA.data());
		Real *__restrict__ FA_data=reinterpret_cast<Real*>(FA.data());


		// timer[13] -= MPI_Wtime();
		for (size_t c=0; c<num_p_cols; c++)
			for (size_t ix=0; ix<maxix; ix++)
				std::copy(IA_data + ix*plane_size_YX + c*copy_size_YX, IA_data + ix*plane_size_YX + (c+1)*copy_size_YX, buffer_Y + c*block_size_YX + ix*copy_size_YX);
		// timer[13] += MPI_Wtime();

		// timer[14] -= MPI_Wtime();
		MPI_Alltoall(buffer_Y, block_size_YX, MPI_Real, FA_data, block_size_YX, MPI_Real, MPI_COMM_ROW);
		// timer[14] += MPI_Wtime();
	}
	else
		FA=Array<Complex,3>(reinterpret_cast<Complex*>(IA.data()), FA_shape, neverDeleteData);

}

void FFTK::Transpose_YZ(Array<Complex,3> IA, Array<Real,3> RA)
{
	if (num_p_rows>1) {
		Real *__restrict__ IA_data=reinterpret_cast<Real*>(IA.data());
		Real *__restrict__ Ar_data=reinterpret_cast<Real*>(RA.data());

		// timer[15] -= MPI_Wtime();
		for (size_t r=0; r<num_p_rows; r++)
			for (size_t ix=0; ix<maxix; ix++)
				std::copy(IA_data + ix*plane_size_YZ + r*copy_size_YZ, IA_data + ix*plane_size_YZ + (r+1)*copy_size_YZ, buffer_Y + r*block_size_YZ + ix*copy_size_YZ);
		// timer[15] += MPI_Wtime();


		// timer[16] -= MPI_Wtime();
		MPI_Alltoall(buffer_Y, block_size_YZ, MPI_Real, buffer_Z, block_size_YZ, MPI_Real, MPI_COMM_COL);
		// timer[16] += MPI_Wtime();


		// timer[17] -= MPI_Wtime();
		for (size_t r=0; r<num_p_rows; r++)
			for (size_t j=0; j<maxry*maxrx; j++)
				std::copy(buffer_Z + r*block_size_YZ + j*Iz_unit*maxiz,buffer_Z + r*block_size_YZ + (j+1)*Iz_unit*maxiz, Ar_data + j*Rz + r*Iz_unit*maxiz);
		// timer[17] += MPI_Wtime();
	}
	else 
		RA=Array<Real,3>(reinterpret_cast<Real*>(IA.data()), RA_shape, neverDeleteData);


 	if (Iz_unit*maxiz*num_p_rows < Rz) {
		// timer[18] -= MPI_Wtime();
		Zero_pad_last_plane(RA);
		// timer[18] += MPI_Wtime();
	}

}

void FFTK::Transpose_ZY(Array<Real,3> RA, Array<Complex,3> IA)
{
	if (num_p_rows>1) {
		Real *__restrict__ RA_data=reinterpret_cast<Real*>(RA.data());
		Real *__restrict__ IA_data=reinterpret_cast<Real*>(IA.data());

		// timer[19] -= MPI_Wtime();
		for (size_t r=0; r<num_p_rows; r++)
			for (size_t j=0; j<maxry*maxrx; j++)
				std::copy(RA_data + j*Rz + r*Iz_unit*maxiz, RA_data + j*Rz + (r+1)*Iz_unit*maxiz, buffer_Z + r*block_size_YZ + j*Iz_unit*maxiz);
		// timer[19] += MPI_Wtime();

		// timer[20] -= MPI_Wtime();
		MPI_Alltoall(buffer_Z, block_size_YZ, MPI_Real, buffer_Y, block_size_YZ, MPI_Real, MPI_COMM_COL);
		// timer[20] += MPI_Wtime();

		// timer[21] -= MPI_Wtime();
		for (size_t r=0; r<num_p_rows; r++)
			for (size_t j=0; j<maxix; j++)
				std::copy(buffer_Y + r*block_size_YZ + j*copy_size_YZ, buffer_Y + r*block_size_YZ + (j+1)*copy_size_YZ, IA_data + j*plane_size_YZ + r*copy_size_YZ);
		// timer[21] += MPI_Wtime();
	}
	else
		IA=Array<Complex,3>(reinterpret_cast<Complex*>(RA.data()), FA_shape, neverDeleteData);

}

void FFTK::Zero_pad_last_plane(Array<Real,2> Ar) {
	Ar(Range::all(),Range(Nz,Nz+1))=0;
}

void FFTK::Zero_pad_last_plane(Array<Real,3> Ar) {
	Ar(Range::all(),Range::all(),Range(Nz,Nz+1))=0;
}

void  FFTK::Normalize(Array<Complex,2> A) {
	A*=normalizing_factor;
}

void  FFTK::Normalize(Array<Complex,3> A) {
	A*=normalizing_factor;
}


//After SinCos transform, shift right
void FFTK::ArrayShiftRight_basic(void *data, TinyVector<int,3> shape, int axis)
{	
	Array<Real,3> Ar((Real*)(data), shape, neverDeleteData);

	if (axis == 'X') {
		Ar(Range(Ar.extent(0)-1,1,-1),Range::all(),Range::all()) = Ar(Range(Ar.extent(0)-2,0,-1),Range::all(),Range::all());
		Ar(0,Range::all(),Range::all()) = 0.0;
	}
	else if (axis == 'Y') {
		Ar(Range::all(),Range(Ar.extent(1)-1,1,-1),Range::all()) = Ar(Range::all(),Range(Ar.extent(1)-2,0,-1),Range::all());
		Ar(Range::all(),0,Range::all()) = 0.0;	
	}
	else if (axis == 'Z') {
		Ar(Range::all(),Range::all(),Range(Ar.extent(2)-1,1,-1)) = Ar(Range::all(),Range::all(),Range(Ar.extent(2)-2,0,-1));
		Ar(Range::all(),Range::all(),0) = 0.0;	
	}
}

void FFTK::ArrayShiftRight(Array<Real,3> Ar, int axis)
{	
	ArrayShiftRight_basic(Ar.data(), Ar.shape(), axis);
}

void FFTK::ArrayShiftRight(Array<Complex,3> A, int axis)
{	
	ArrayShiftRight_basic(A.data(), A.shape()*shape(1,1,2), axis);
}

void FFTK::ArrayShiftRight(Array<Real,2> Ar, int axis)
{	
	ArrayShiftRight_basic(Ar.data(), shape(Ar.extent(0), 1, Ar.extent(1)), axis);
}

void FFTK::ArrayShiftRight(Array<Complex,2> A, int axis)
{	
	ArrayShiftRight_basic(A.data(), shape(A.extent(0), 1, A.extent(1)*2), axis);
}

//Before inverse SinCos transform, shift left
void FFTK::ArrayShiftLeft_basic(void *data, TinyVector<int,3> shape, int axis)
{
	Array<Real,3> Ar((Real*)(data), shape, neverDeleteData);

	if (axis == 'X') {
		Ar(Range(0,Ar.extent(0)-2),Range::all(),Range::all()) = Ar(Range(1,Ar.extent(0)-1),Range::all(),Range::all());
		Ar(Ar.extent(0)-1,Range::all(),Range::all()) = 0.0;	
	}
	else if (axis == 'Y') {
		Ar(Range::all(),Range(0,Ar.extent(1)-2),Range::all()) = Ar(Range::all(),Range(1,Ar.extent(1)-1),Range::all());
		Ar(Range::all(),Ar.extent(1)-1,Range::all()) = 0.0;

	}
	else if (axis == 'Z') {
		Ar(Range::all(),Range::all(),Range(0,Ar.extent(2)-2)) = Ar(Range::all(),Range::all(),Range(1,Ar.extent(2)-1));
		Ar(Range::all(),Range::all(),Ar.extent(2)-1) = 0.0;	
	}
}

void FFTK::ArrayShiftLeft(Array<Real,3> Ar, int axis)
{	
	ArrayShiftLeft_basic(Ar.data(), Ar.shape(), axis);
}

void FFTK::ArrayShiftLeft(Array<Complex,3> A, int axis)
{	
	ArrayShiftLeft_basic(A.data(), A.shape()*shape(1,1,2), axis);
}

void FFTK::ArrayShiftLeft(Array<Real,2> Ar, int axis)
{	
	ArrayShiftLeft_basic(Ar.data(), shape(Ar.extent(0), 1, Ar.extent(1)), axis);
}

void FFTK::ArrayShiftLeft(Array<Complex,2> A, int axis)
{	
	ArrayShiftLeft_basic(A.data(), shape(A.extent(0), 1, A.extent(1)*2), axis);
}

void FFTK::To_slab(Array<Real,3> ArPencil, Array<Real,3> ArSlab)
{
	/*MPI_Gather(ArPencil.data(), ArPencil.size(), MPI_Real,
           ArSlab.data(), 1, MPI_Vector_resized_ZY_planes, 0,
           MPI_COMM_COL);
	*/

	Array<Real,3> ArSlab_temp(num_p_rows*maxrx,Ry/num_p_rows,Rz);

	MPI_Gather(ArPencil.data(), ArPencil.size(), MPI_Real,
           ArSlab_temp.data(), ArPencil.size(), MPI_Real, 0,
           MPI_COMM_COL);


	if (my_row_id==0) {
		for (int p=0; p<num_p_rows; p++) {
			ArSlab(Range::all(), Range(p*Ry/num_p_rows,(p+1)*Ry/num_p_rows-1), Range::all())
				= ArSlab_temp(Range(p*maxrx,(p+1)*maxrx-1),Range::all(),Range::all());
		}
	}
}

void FFTK::To_pencil(Array<Real,3> ArSlab, Array<Real,3> ArPencil)
{
	Array<Real,3> ArSlab_temp(num_p_rows*maxrx,Ry/num_p_rows,Rz);

	if (my_row_id==0) {
		for (int p=0; p<num_p_rows; p++) {
			ArSlab_temp(Range(p*maxrx,(p+1)*maxrx-1),Range::all(),Range::all())
				= ArSlab(Range::all(), Range(p*Ry/num_p_rows,(p+1)*Ry/num_p_rows-1), Range::all());
		}
	}

	MPI_Scatter(ArSlab_temp.data(), ArPencil.size(), MPI_Real,
		ArPencil.data(), ArPencil.size(), MPI_Real, 0,
        MPI_COMM_COL);
}
MPI_Comm FFTK::Get_communicator(string which) {
	if (which=="ROW")
		return MPI_COMM_ROW;
	else if (which=="COL")
		return MPI_COMM_COL;
	else {
		if (my_id==0)
			cerr << "Invalid communicator: " << which << endl;
		MPI_Finalize();
		MPI_Abort(MPI_COMM_WORLD, 0);
	}
}


TinyVector<int,3> FFTK::Get_FA_shape()
{
	return FA_shape;
}
TinyVector<int,3> FFTK::Get_IA_shape()
{
	return IA_shape;
}
TinyVector<int,3> FFTK::Get_RA_shape()
{
	return RA_shape;
}

TinyVector<int,3> FFTK::Get_FA_start()
{
	return TinyVector<int, 3>(fx_start,fy_start,fz_start);
}
TinyVector<int,3> FFTK::Get_IA_start()
{
	return TinyVector<int, 3>(ix_start,iy_start,iz_start);
}
TinyVector<int,3> FFTK::Get_RA_start()
{
	return TinyVector<int, 3>(rx_start,ry_start,rz_start);
}

int FFTK::Get_row_id()
{
	return my_row_id;
}
int FFTK::Get_col_id()
{
	return my_col_id;
}

int FFTK::Get_num_p_rows()
{
	return num_p_rows;
}
int FFTK::Get_num_p_cols()
{
	return num_p_cols;
}

void FFTK::Finalize() {
	delete[] buffer_Y;
	delete[] buffer_Z;

	FFTW_DESTROY_PLAN(fft_plan_Fourier_forward_x);
	FFTW_DESTROY_PLAN(fft_plan_Fourier_inverse_x);
	FFTW_DESTROY_PLAN(fft_plan_Sine_forward_x);
	FFTW_DESTROY_PLAN(fft_plan_Sine_inverse_x);
	FFTW_DESTROY_PLAN(fft_plan_Cosine_forward_x);
	FFTW_DESTROY_PLAN(fft_plan_Cosine_inverse_x);
	FFTW_DESTROY_PLAN(fft_plan_Fourier_forward_y);
	FFTW_DESTROY_PLAN(fft_plan_Fourier_inverse_y);
	FFTW_DESTROY_PLAN(fft_plan_Sine_forward_y);
	FFTW_DESTROY_PLAN(fft_plan_Sine_inverse_y);
	FFTW_DESTROY_PLAN(fft_plan_Cosine_forward_y);
	FFTW_DESTROY_PLAN(fft_plan_Cosine_inverse_y);
	FFTW_DESTROY_PLAN(fft_plan_Fourier_forward_z);
	FFTW_DESTROY_PLAN(fft_plan_Fourier_inverse_z);
	FFTW_DESTROY_PLAN(fft_plan_Sine_forward_z);
	FFTW_DESTROY_PLAN(fft_plan_Sine_inverse_z);
	FFTW_DESTROY_PLAN(fft_plan_Cosine_forward_z);
	FFTW_DESTROY_PLAN(fft_plan_Cosine_inverse_z);
}
