#ifndef _FFTK_H
#define _FFTK_H

#include <mpi.h>
#include <complex>
#include <fftw3-mpi.h>

#include <blitz/array.h>

#ifdef FFTK_ENABLE_THREADS
	#include <omp.h>
#endif


using namespace std;
using namespace blitz;


//*********************************************************************************************

// Defining REAL_DOUBLE: switch for setting double
#if defined(REAL_DOUBLE)

#define Real                            double
#define MPI_Real                        MPI_DOUBLE
#define FFTW_Complex                    fftw_complex

#define FFTW_MPI_INIT                   fftw_mpi_init

#define FFTW_PLAN                       fftw_plan
#define FFTW_MPI_PLAN_DFT_3D            fftw_mpi_plan_dft_3d   // for GP
#define FFTW_MPI_PLAN_DFT_R2C_2D        fftw_mpi_plan_dft_r2c_2d
#define FFTW_MPI_PLAN_DFT_C2R_2D        fftw_mpi_plan_dft_c2r_2d
#define FFTW_MPI_PLAN_DFT_R2C_3D        fftw_mpi_plan_dft_r2c_3d
#define FFTW_MPI_PLAN_DFT_C2R_3D        fftw_mpi_plan_dft_c2r_3d
#define FFTW_PLAN_MANY_R2R              fftw_plan_many_r2r
#define FFTW_PLAN_MANY_DFT              fftw_plan_many_dft
#define FFTW_PLAN_MANY_DFT_R2C          fftw_plan_many_dft_r2c
#define FFTW_PLAN_MANY_DFT_C2R          fftw_plan_many_dft_c2r

#define FFTW_EXECUTE                    fftw_execute
#define FFTW_EXECUTE_R2R                fftw_execute_r2r
#define FFTW_EXECUTE_DFT                fftw_execute_dft
#define FFTW_EXECUTE_DFT_R2C            fftw_execute_dft_r2c
#define FFTW_EXECUTE_DFT_C2R            fftw_execute_dft_c2r
#define FFTW_MPI_EXECUTE_DFT            fftw_mpi_execute_dft
#define FFTW_MPI_EXECUTE_DFT_R2C        fftw_mpi_execute_dft_r2c        
#define FFTW_MPI_EXECUTE_DFT_C2R        fftw_mpi_execute_dft_c2r  
#define FFTW_DESTROY_PLAN               fftw_destroy_plan


// Define REAL_FLOAT: switch for setting float
#elif defined(REAL_FLOAT)

#define Real                            float
#define MPI_Real                        MPI_FLOAT
#define FFTW_Complex                    fftwf_complex

#define FFTW_MPI_INIT                   fftwf_mpi_init

#define FFTW_PLAN                       fftwf_plan
#define FFTW_MPI_PLAN_DFT_3D            fftwf_mpi_plan_dft_3d   // for GP
#define FFTW_MPI_PLAN_DFT_R2C_2D        fftwf_mpi_plan_dft_r2c_2d
#define FFTW_MPI_PLAN_DFT_C2R_2D        fftwf_mpi_plan_dft_c2r_2d
#define FFTW_MPI_PLAN_DFT_R2C_3D        fftwf_mpi_plan_dft_r2c_3d
#define FFTW_MPI_PLAN_DFT_C2R_3D        fftwf_mpi_plan_dft_c2r_3d
#define FFTW_PLAN_MANY_R2R              fftwf_plan_many_r2r
#define FFTW_PLAN_MANY_DFT              fftwf_plan_many_dft
#define FFTW_PLAN_MANY_DFT_R2C          fftwf_plan_many_dft_r2c
#define FFTW_PLAN_MANY_DFT_C2R          fftwf_plan_many_dft_c2r

#define FFTW_EXECUTE                    fftwf_execute
#define FFTW_EXECUTE_R2R                fftwf_execute_r2r
#define FFTW_EXECUTE_DFT                fftwf_execute_dft
#define FFTW_EXECUTE_DFT_R2C            fftwf_execute_dft_r2c
#define FFTW_EXECUTE_DFT_C2R            fftwf_execute_dft_c2r
#define FFTW_MPI_EXECUTE_DFT            fftwf_mpi_execute_dft
#define FFTW_MPI_EXECUTE_DFT_R2C        fftwf_mpi_execute_dft_r2c        
#define FFTW_MPI_EXECUTE_DFT_C2R        fftwf_mpi_execute_dft_c2r        
#define FFTW_DESTROY_PLAN               fftwf_destroy_plan

#endif


#ifndef Complex
#define Complex  complex<Real>
#endif

// Switches for FFTW plans 

#if defined(ESTIMATE)
#define FFTW_PLAN_FLAG FFTW_ESTIMATE

#elif defined(MEASURE)
#define FFTW_PLAN_FLAG FFTW_MEASURE

#elif defined(PATIENT)
#define FFTW_PLAN_FLAG FFTW_PATIENT

#elif defined(EXHAUSTIVE)
#define FFTW_PLAN_FLAG FFTW_EXHAUSTIVE
#endif


#define IN_ALL_PROCS(l, expr) \
for (int l=0; l<numprocs; l++) { \
    if (my_id==l)\
        {expr;}\
    MPI_Barrier(MPI_COMM_WORLD);\
}



class FFTK {
private:
	string basis;

	int my_id;
	int numprocs;

	MPI_Comm MPI_COMM_ROW;
	MPI_Comm MPI_COMM_COL;

	int my_row_id;
	int my_col_id;

	int Nx,Ny,Nz;

	int Fx,Fy,Fz;
	int Ix,Iy,Iz;
	int Rx,Ry,Rz;

	int Fz_unit, Iz_unit, Rz_unit;

	int num_p_rows;
	int num_p_cols;

	//Set Array size
	TinyVector<int,3> FA_shape, IA_shape, RA_shape;

	int maxfx, maxfy, maxfz;
	int maxix, maxiy, maxiz;
	int maxrx, maxry, maxrz;

	int fx_start, fy_start, fz_start;
	int ix_start, iy_start, iz_start;
	int rx_start, ry_start, rz_start;

	FFTW_PLAN fft_plan_Fourier_forward_x;
	FFTW_PLAN fft_plan_Fourier_inverse_x;
	FFTW_PLAN fft_plan_Fourier_forward_y;
	FFTW_PLAN fft_plan_Fourier_inverse_y;
	FFTW_PLAN fft_plan_Fourier_forward_z;
	FFTW_PLAN fft_plan_Fourier_inverse_z;

	FFTW_PLAN fft_plan_Sine_forward_x;
	FFTW_PLAN fft_plan_Sine_inverse_x;
	FFTW_PLAN fft_plan_Sine_forward_y;
	FFTW_PLAN fft_plan_Sine_inverse_y;
	FFTW_PLAN fft_plan_Sine_forward_z;
	FFTW_PLAN fft_plan_Sine_inverse_z;

	FFTW_PLAN fft_plan_Cosine_forward_x;
	FFTW_PLAN fft_plan_Cosine_inverse_x;
	FFTW_PLAN fft_plan_Cosine_forward_y;
	FFTW_PLAN fft_plan_Cosine_inverse_y;
	FFTW_PLAN fft_plan_Cosine_forward_z;
	FFTW_PLAN fft_plan_Cosine_inverse_z;

	Real *buffer_Y;
	Real *buffer_Z;

	size_t plane_size_YX;
	size_t copy_size_YX;
	size_t block_size_YX;

	size_t plane_size_YZ;
	size_t block_size_YZ;
	size_t copy_size_YZ;
	//


	Array<Complex,3> FA;
	Array<Complex,3> IA;
	Array<Real,3> RA;

	Real normalizing_factor;

	void set_vars();
	void init_transpose();
	void init_transform();
	void init_timers();
	void Transpose_XY(Array<Complex,3> A1, Array<Complex,3> A2);
	void Transpose_YX(Array<Complex,3> A1, Array<Complex,3> A2);
	void Transpose_YZ(Array<Complex,3> A, Array<Real,3> Ar);
	void Transpose_ZY(Array<Real,3> Ar, Array<Complex,3> A);

	void ArrayShiftRight(Array<Real,2> Ar, int axis);
	void ArrayShiftRight(Array<Complex,2> A, int axis);
	void ArrayShiftLeft(Array<Real,2> Ar, int axis);
	void ArrayShiftLeft(Array<Complex,2> A, int axis);

	void ArrayShiftRight_basic(void *data, TinyVector<int,3> shape, int axis);
	void ArrayShiftRight(Array<Real,3> Ar, int axis);
	void ArrayShiftRight(Array<Complex,3> A, int axis);
	void ArrayShiftLeft_basic(void *data, TinyVector<int,3> shape, int axis);
	void ArrayShiftLeft(Array<Real,3> Ar, int axis);
	void ArrayShiftLeft(Array<Complex,3> A, int axis);

public:
	double *timer;

	void Init(string basis, int Nx, int Ny, int Nz, int num_p_rows);
	void Finalize();
	
	void Forward_transform(string basis_option, Array<Real,2> RA, Array<Complex,2> FA);
	void Inverse_transform(string basis_option, Array<Complex,2> FA, Array<Real,2> RA);

	void Forward_transform(string basis_option, Array<Real,3> RA, Array<Complex,3> FA);
	void Inverse_transform(string basis_option, Array<Complex,3> FA, Array<Real,3> RA);

	void Transpose(Array<Real,2> RA, Array<Complex,2> FA);
	void Transpose(Array<Complex,2> FA, Array<Real,2> RA);
	
	void Transpose(Array<Real,3> Ar, Array<Complex,3> A);
	void Transpose(Array<Complex,3> A, Array<Real,3> Ar);
	
	void Zero_pad_last_plane(Array<Real,2> Ar);
	void Zero_pad_last_plane(Array<Real,3> Ar);

	void Normalize(Array<Complex,2> A);
	void Normalize(Array<Complex,3> A);

	void To_slab(Array<Real,3> ArPencil, Array<Real,3> ArSlab);
	void To_pencil(Array<Real,3> ArSlab, Array<Real,3> ArPencil);

	MPI_Comm Get_communicator(string which);

	TinyVector<int,3> Get_FA_shape();
	TinyVector<int,3> Get_IA_shape();
	TinyVector<int,3> Get_RA_shape();

	TinyVector<int,3> Get_FA_start();
	TinyVector<int,3> Get_IA_start();
	TinyVector<int,3> Get_RA_start();

	int Get_row_id();
	int Get_col_id();

	int Get_num_p_rows();
	int Get_num_p_cols();

};

#endif
